// V-E Cursor Demo (VECC version)
// Virtual-E (c) 1999-2000 <virtuale98@yahoo.com>
// Originally coded by Alberto Covarrubias
//
// Modified by VeeBee Development (http://members.lycos.co.uk/veebeedev/)
// Updated & made 'interactive' by David Williamson (lameboyadvance)
//                                     (david.williamson@excite.com)

// Demonstrates how to create a basic VB program using
// Alberto Covarrubias' excellent Virtual-E C Compiler
//
// Shows how to:
// * Define chars           * Manipulate display registers
// * Reset the VB           * Set worlds
// * Init the display       * Manipulate worlds
// * Copy data to memory    * Read the controller
//
// (NOTE: You can probably see that this uses both hexadecimal (0x##),
// and binary (0/1) numbers. If you don't know how to use them, learn.)
//
// I really suggest reading background material like the 'VB Programmer's
// Manual' before attempting to make something for the VB.
// It helps clear a lot of things up.

#include "vbaddr.h"  // Define the known VB memory locations
#include "vbtypes.h" // Define the data types used in a VB
#include "vbhw.h"    // Define the common methods built in to VECC

// The following is taken from the VB programmer's manual:
// A Char is an individual 8x8 pixel, 4 colour bitmap, and is the
// basic element upon which all VB graphics are based.
// Both BGMaps and OBJs are collections of Chars.
//
// Each of the following lines contain a different char. Each hex number contains
// a row in that char. Pixel rows are stored in a right to left format.
// For each pixel:
// 11 = white
// 10 = light grey
// 01 = dark grey
// 00 = black/clear
// These colors can be swapped around using palettes. More on this below.
//
// Define each char into a HWORD array
static HWORD chCursor[] = {
    0x003C, 0x03EC, 0x3EAC, 0xE96F, 0xE96F, 0x3EAC, 0x03EC, 0x003C, // R Arrow
    0x3C00, 0x3BC0, 0x3ABC, 0xF96B, 0xF96B, 0x3ABC, 0x3BC0, 0x3C00, // L Arrow
    0x03C0, 0x0EB0, 0x0EB0, 0x396C, 0x396C, 0xEAAB, 0xFFFF, 0x03C0, // U Arrow
    0x03C0, 0xFFFF, 0xEAAB, 0x396C, 0x396C, 0x0EB0, 0x0EB0, 0x03C0, // D Arrow
    0x0000, 0x3FFC, 0x3AAC, 0x396C, 0x396C, 0x3AAC, 0x3FFC, 0x0000, // + Arrow
    0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3FFC, 0x0000, // 'U'
    0x0FF0, 0x3C3C, 0x00FC, 0x0FC0, 0x3F00, 0x3C3C, 0x0FF0, 0x0000, // 'S'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0xC0F0, 0xFFFC, 0x0000, // 'E'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x3FFC, 0x33CC, 0x03C0, 0x03C0, 0x03C0, 0x03C0, 0x0FF0, 0x0000, // 'T'
    0x3C3C, 0x3C3C, 0x3C3C, 0x3FFC, 0x3C3C, 0x3C3C, 0x3C3C, 0x0000, // 'H'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0xC0F0, 0xFFFC, 0x0000, // 'E'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x03FC, 0x00F0, 0x00F0, 0x00F0, 0xC0F0, 0xF0F0, 0xFFFC, 0x0000, // 'L'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0xC0F0, 0xFFFC, 0x0000, // 'E'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0x00F0, 0x03FC, 0x0000, // 'F'
    0x3FFC, 0x33CC, 0x03C0, 0x03C0, 0x03C0, 0x03C0, 0x0FF0, 0x0000, // 'T'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x0FFC, 0x3CF0, 0x3CF0, 0x0FF0, 0x00F0, 0x00F0, 0x03FC, 0x0000, // 'P'
    0x03C0, 0x0FF0, 0x3C3C, 0x3C3C, 0x3FFC, 0x3C3C, 0x3C3C, 0x0000, // 'A'
    0x3FFC, 0x3CF0, 0xF0F0, 0xF0F0, 0xF0F0, 0x3CF0, 0x3FFC, 0x0000, // 'D'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x3FFC, 0x33CC, 0x03C0, 0x03C0, 0x03C0, 0x03C0, 0x0FF0, 0x0000, // 'T'
    0x0FC0, 0x3CF0, 0xF03C, 0xF03C, 0xF03C, 0x3CF0, 0x0FC0, 0x0000, // 'O'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0xF03C, 0xFCFC, 0xFFFC, 0xF33C, 0xF03C, 0xF03C, 0xF03C, 0x0000, // 'M'
    0x0FC0, 0x3CF0, 0xF03C, 0xF03C, 0xF03C, 0x3CF0, 0x0FC0, 0x0000, // 'O'
    0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x0FF0, 0x03C0, 0x0000, // 'V'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0xC0F0, 0xFFFC, 0x0000, // 'E'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x3FFC, 0x33CC, 0x03C0, 0x03C0, 0x03C0, 0x03C0, 0x0FF0, 0x0000, // 'T'
    0x3C3C, 0x3C3C, 0x3C3C, 0x3FFC, 0x3C3C, 0x3C3C, 0x3C3C, 0x0000, // 'H'
    0xFFFC, 0xC0F0, 0x0CF0, 0x0FF0, 0x0CF0, 0xC0F0, 0xFFFC, 0x0000, // 'E'
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, // ' '
    0x3FC0, 0xF0F0, 0x003C, 0x003C, 0x003C, 0xF0F0, 0x3FC0, 0x0000, // 'C'
    0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3C3C, 0x3FFC, 0x0000, // 'U'
    0x3FFC, 0xF0F0, 0xF0F0, 0x3FF0, 0x0FF0, 0x3CF0, 0xF0FC, 0x0000, // 'R'
    0x0FF0, 0x3C3C, 0x00FC, 0x0FC0, 0x3F00, 0x3C3C, 0x0FF0, 0x0000, // 'S'
    0x0FC0, 0x3CF0, 0xF03C, 0xF03C, 0xF03C, 0x3CF0, 0x0FC0, 0x0000, // 'O'
    0x3FFC, 0xF0F0, 0xF0F0, 0x3FF0, 0x0FF0, 0x3CF0, 0xF0FC, 0x0000  // 'R'
};

// The following is taken from the VB programmer's manual:
// BGMaps are like checker boards made of 64x64 Chars (512x512 pixels).
// Since one can 'move' the whole BGMap relative to it's World, but not
// move the individual Chars relative to each other, they are best at
// displaying large scenes like background graphics.
//
// Each number indicates what char will be displayed in the BGMap.
//     (eg. 0, 1, 63, 64, 65)
//     1st 'checker board' spot on the BGMap (0,0) will contain Char 0
//     2nd 'checker board' spot on the BGMap (1,0) will contain Char 1
//     3rd 'checker board' spot on the BGMap (2,0) will contain Char 63
//     4th 'checker board' spot on the BGMap (3,0) will contain Char 64
//     5th 'checker board' spot on the BGMap (4,0) will contain Char 65
//     And so on...
//
// Define the char index
static HWORD bgCursor[] = {
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39
};

// Init the display
// Uses the VIP (display) registers to reset all
// the BGMap and OBJ palettes and colours
void DisplayInit()
{
    // The colour brightness is in the following format:
    // from 0 (black/off) up to 0x80 (full brightness)
    VIP_REGS(0x24) = 0;    // Colour 1 brightness (BRTA)
    VIP_REGS(0x26) = 0;    // Colour 2 brightness (BRTB)
    VIP_REGS(0x28) = 0;    // Colour 3 brightness (BRTC)
    
    // The map/object palettes are in the following format:
    // Col 3, Col 2, Col 1, Col 0  (Usually 11 10 01 00)
    VIP_REGS(0x60) = 0xE4; // BGMap Palette 0     (GPLT0)
    VIP_REGS(0x62) = 0xE4; // BGMap Palette 1     (GPLT1)
    VIP_REGS(0x64) = 0xE4; // BGMap Palette 2     (GPLT2)
    VIP_REGS(0x65) = 0xE4; // BGMap Palette 3     (GPLT3)
    VIP_REGS(0x66) = 0xE4; // OBJ Palette 0       (JPLT0)
    VIP_REGS(0x6A) = 0xE4; // OBJ Palette 1       (JPLT1)
    VIP_REGS(0x6C) = 0xE4; // OBJ Palette 2       (JPLT2)
    VIP_REGS(0x6E) = 0xE4; // OBJ Palette 3       (JPLT3)
    
    // The bg colours are in the same format as the palettes
    VIP_REGS(0x70) = 0;    // Background colour   (BKCOL)
}

// Displays and moves the cursor
// First sets up the world used, then turns on the display,
// then sets the brightness and palettes, and finally creates
// an infinite loop in which the controller input is read, and
// is used to determine the current status of the cursor.
void DisplayCursor()
{
    // The following is taken from the VB programmer's manual:
    // Worlds are collections of OBJs and BGMaps that are layered and
    // clipped to fit each screen (384x224).
    // There are a total of 32 Worlds, but not all 32 Worlds need be used.
    // Worlds are drawn last to first, starting at 31 and ending with 0. Only
    // Worlds with at least one visible bit set (LON or RON = 1) are drawn.
    // However, if a World is marked as the END World, it and all remaining
    // Worlds are skipped.
    // As the VB renders the left and right screens, parallax is factored in,
    // by subtracting GP from GX for the left screen, and adding GP to GX for
    // the right screen. Any special effects, like sprite scaling, and rotation,
    // are also rendered at this time.
    // (NOTE: The worlds with the lowest number are displayed last.
    //     ie. World 30 is displayed 'on top of' World 31)
    //
    // vbSetWorld(HEADER, GX, GP, GY, MX, MP, MY, W, H) structure:
    // HEADER: 16 bits
    //     F    E   D C  B A  9 8   7    6   5  4  3  2  1  0
    //    LON  RON  BGM  SCX  SCY  OVR  END  0  0  BGMAP_BASE
    //             (eg. 0xC000 = LON+RON, 0x0040 = END World)
    // GX (0x000  - 0x17F)
    // GP (-0x100 - 0x0FF)
    // GY (0x000  - 0x0DF)
    // MX (0x000  - 0xFFF)
    // MP (-0x100 - 0x0FF)
    // MY (0x000  - 0xFFF)
    // W  (0x000  - 0xFFF)
    // H  (0x000  - 0xFFF)
    //
    // LON - Displayed on left screen
    // RON - Displayed on right screen
    // BGM - World type: (00)=BGMap, (01)=HBias, (10)=Affine, (11)=OBJ Collection.
    // SCX, SCY - How the chars are tiled. I haven't learned how to use this yet.
    // OVR - Overplane? I don't know how to use this one either.
    // END - This and any remaining Worlds should not be drawn.
    // BGMAP_BASE - The number of the first BGMap to display.
    // GX - Screen X start position.
    // GY - Screen Y start position.
    // GP - Parallax offset for screen X position, true X coordinates (TX) are
    // computed by: TX(Left) = GX - GP and TX(Right) = GX + GP.
    //             (eg. 2 = each screen is offset 2 pixels & appears to 'pop out')
    // MX - Buffer X start position. (X coord to start reading image)
    // MY - Buffer Y start position. (Y coord to start reading image)
    // MP - Parallax offset for Buffer X position, actually shifts the start
    // address to be cut out, in the X direction, to make a "Window" effect.
    // Each eye sees a touch more on the edges than the other.
    // W - Width to cut out from the buffer and paste on the screen.
    // H - Height to cut out from the buffer and paste on the screen.
    // Must be a minimum of 8 pixels high.
    
    int x = 0x60; // Cursor X offset
    int y = 0x6F; // Cursor Y offset
    
    vbSetWorld(31, 0xC000, 0x60 + x, 0, 0x6F + y, 0, 0, 0, 7, 7); // Display cursor
    // World 31 is displayed on both screens. Starts at (0x60+x, 0x6F+y), has 0 parallax,
    // no MX MY or MP, 8 pixels (7 - 1) wide & 8 pixels (7 - 1) high.
    
    vbSetWorld(30, 0xC000, 0x33, 2, 0x60, 40, 0, 0, 279, 7);      // Display text
    // World 30 is displayed on both screens. Starts at (0x33, 0), has 2 parallax, MX
    // (X pos to start reading buffer) starts at 40, no MY or MP, 270 x 8 pixels.
    
    vbSetWorld(29, 0x0040, 0, 0, 0, 0, 0, 0, 0, 0); // Blank world and END bit set
    // to notify the VB of no more worlds to display. END world is not displayed.
    
    vbDisplayOn();         // Turn the VB display on
    VIP_REGS(0x60) = 0xE4; // Set BGMap palette 0     (GPLT0)
    VIP_REGS(0x24) = 0x80; // Set Colour 1 brightness (BRTA)
    VIP_REGS(0x26) = 0x80; // Set Colour 2 brightness (BRTB)
    VIP_REGS(0x28) = 0x80; // Set Colour 3 brightness (BRTC)
    // vbFXFadeIn();       // Fade in (gradually increase brightness)
    
    while(1) // Infinite loop to continuously read controller
    {
        if (vbReadPad() & 0x100)      // L pad right pressed
        {
            if (x < 376)
                x++;
            
            vbSetWorld(31, 0xC000, x, 0, y, 0, 0, 0, 7, 7);  // Display cursor
            // Displayed on both screens. Starts at (x, y), has 0 parallax, no
            // MX MY or MP, 8 x 8 pixels.
            
            vbWaitFrame(1);           // Delay for 1 (millisecond?)
        }
        else if (vbReadPad() & 0x200) // L pad left pressed
        {
            if (x > 0)
                x--;
            
            vbSetWorld(31, 0xC000, x, 0, y, 8, 0, 0, 7, 7);  // Display cursor
            // Displayed on both screens. Starts at (x, y), has 0 parallax, MX (X
            // pos to start reading buffer) starts at 8, no MY or MP, 8 x 8 pixels.
            
            vbWaitFrame(1);           // Delay for 1 (millisecond?)
        }
        else if (vbReadPad() & 0x800) // L pad up pressed
        {
            if (y > 0)
                y--;
            
            vbSetWorld(31, 0xC000, x, 0, y, 16, 0, 0, 7, 7); // Display cursor
            // Displayed on both screens. Starts at (x, y), has 0 parallax, MX (X
            // pos to start reading buffer) starts at 16, no MY or MP, 8 x 8 pixels.
            
            vbWaitFrame(1);           // Delay for 1 (millisecond?)
        }
        else if (vbReadPad() & 0x400) // L pad down pressed
        {
            if (y < 216)
                y++;
            
            vbSetWorld(31, 0xC000, x, 0, y, 24, 0, 0, 7, 7); // Display cursor
            // Displayed on both screens. Starts at (x, y), has 0 parallax, MX (X
            // pos to start reading buffer) starts at 24, no MY or MP, 8 x 8 pixels.
            
            vbWaitFrame(1);           // Delay for 1 (millisecond?)
        }
        else                          // No L pad pressed
            vbSetWorld(31, 0xC000, x, 0, y, 32, 0, 0, 7, 7); // Display cursor
            // Displayed on both screens. Starts at (x, y), has 0 parallax, MX (X
            // pos to start reading buffer) starts at 32, no MY or MP, 8 x 8 pixels.
    }
}

// Main method
// Resets the VB, Inits the display, copies chars to memory, and
// then finally calls the display cursor method.
// (NOTE: VECC only accepts void main(), not int main() or any arguments)
void main()
{
    vbReset();       // Reset the VB
    DisplayInit();   // Init the display
    memcpy((void*) CharSeg0, (void*) chCursor, 40 * 16); // Copy the chars to CharMap 0
    memcpy((void*) BGMap(0), (void*) bgCursor, 40 * 2);  // Copy the char index to BGMap 0
    DisplayCursor(); // Display the cursor
}
